﻿using System;
using System.Collections.Generic;
using System.Linq;
using System.Text;
using System.Drawing;
using System.IO;

namespace _3dAdd
{
    class Program
    {
        static Bitmap image1;
        static int totalVerticesWritten = 1; // in obj format first vertex is indexed "1".
        struct box
        {
            public int x;
            public int y;
            public int z;
            public int r;
        };

        static void Main(string[] args)
        {
            try
            {
                if (args.Length != 2)
                {
                    System.Console.Error.WriteLine("Usage: 3dAdd.exe [input.bmp] [output.obj]");
                    return;
                }
                string inputFile = args[0];
                string outputFile = args[1];

                image1 = new Bitmap(inputFile, true);

                boxMethod(outputFile);
            }
            catch (Exception e)
            {
                System.Console.Error.WriteLine(e.ToString());
                return;
            }
        }

        static void rayTraceMethod(string outputFile)
        {
            int d = image1.Height;
            bool[, ,] verts = new bool[d, d, d];


            /* The first iteration does both the Z-direction raytracing and also
             * instantiates all the values of verts[,,]. */
            for (int x = 0; x < d; x++)
            {
                for (int y = 0; y < d; y++)
                {
                    for (int z = 0; z < d; z++)
                    {
                        if (image1.GetPixel(x, y).R == 255 && image1.GetPixel(x, y).G == 255 && image1.GetPixel(x, y).B == 255)
                        {
                            verts[x, y, z] = false;
                        }
                        else
                        {
                            verts[x, y, z] = true;
                        }
                    }
                }
            }

            Console.WriteLine("Z-Direction Raytracing...");
            /* Second iteration for Y-direction raytracing */
            for (int x = 0; x < d; x++)
            {
                for (int z = 0; z < d; z++)
                {
                    if (image1.GetPixel(x, z).R == 255 && image1.GetPixel(x, z).G == 255 && image1.GetPixel(x, z).B == 255)
                    {
                        for (int y = 0; y < d; y++)
                        {
                            verts[x, y, z] = false;
                        }
                    }
                }
            }

            Console.WriteLine("Y-Direction Raytracing...");
            /* Finally X-direction raytracing */
            for (int y = 0; y < d; y++)
            {
                for (int z = 0; z < d; z++)
                {
                    if (image1.GetPixel(z, y).R == 255 && image1.GetPixel(z, y).G == 255 && image1.GetPixel(z, y).B == 255)
                    {
                        for (int x = 0; x < d; x++)
                        {
                            verts[x, y, z] = false;
                        }
                    }
                }
            }

            Console.WriteLine("X-Direction Raytracing...");
            /* Ok now we've got all our data formatted so let's output that to a .obj file that we can view in Blender */
            StreamWriter fout = new StreamWriter(outputFile);
            for (int x = 0; x < d; x++)
            {
                for (int y = 0; y < d; y++)
                {
                    for (int z = 0; z < d; z++)
                    {
                        if (verts[x, y, z])
                        {
                            fout.WriteLine("v {0} {1} {2}", x, y, z);
                        }
                    }
                }
            }
            fout.Close();
        }

        static void boxMethod(string of)
        {
            int d = image1.Height;
            bool[, ,] v = new bool[d, d, d];
            StreamWriter fout = new StreamWriter(of);

            for (int y = 0; y < d; y++)
            {
                for (int x = 0; x < d; x++)
                {
                    for (int z = 0; z < d; z++)
                    {
                        // takes a shortcut by just checking R values. Assume black/white image input.
                        v[x, y, z] = (image1.GetPixel(x, y).R == 0 && image1.GetPixel(z, y).R == 0);
                    }
                }
            }

            for (int x = 0; x < d; x++)
            {
                for (int y = 0; y < d; y++)
                {
                    for (int z = 0; z < d; z++)
                    {
                        if (v[x, y, z])
                        {
                            // if anything is zero or d you don't need to do crazy bounds checking
                            if (x == 0 || y == 0 || z == 0 || x == d - 1 || y == d - 1 || z == d - 1)
                            {
                                writeCube(fout, x, y, z, 1);
                            }
                            // This horrible statement sees if there's a point at the location given.
                            // If so, check the six surrounding points and deteremine if any of them are false.
                            // If one of those surrounding points is false, then we need to draw this point.
                            else if (!(v[x - 1, y, z] && v[x + 1, y, z] && v[x, y - 1, z] && v[x, y + 1, z] && v[x, y, z - 1] && v[x, y, z + 1]))
                            {
                                writeCube(fout, x, y, z, 1);
                            }
                        }
                    }
                }
            }

            fout.Close();
        }

        /* Doesn't work, total disaster of wasted effort. */
        static void cubeMethod(string of)
        {
            int d = image1.Height;
            bool[, ,] v = new bool[d, d, d];
            StreamWriter fout = new StreamWriter(of);
            List<box> optimizedBoxes = new List<box>();

            for (int y = 0; y < d; y++)
            {
                for (int x = 0; x < d; x++)
                {
                    for (int z = 0; z < d; z++)
                    {
                        // takes a shortcut by just checking R values. Assume black/white image input.
                        v[x, y, z] = (image1.GetPixel(x, y).R == 0 && image1.GetPixel(z, y).R == 0);
                    }
                }
            }

            for (int x = 0; x < d; x++)
            {
                for (int y = 0; y < d; y++)
                {
                    for (int z = 0; z < d; z++)
                    {
                        bool isCube = true;
                        int r = 0;
                        // Figure out how large our cube can be
                        do
                        {
                            for (int wx = x; wx < x + r && isCube && wx < d; wx++)
                            {
                                for (int wy = y; wy < y + r && isCube && wy < d; wy++)
                                {
                                    for (int wz = z; wz < z + r && isCube && wz < d; wz++)
                                    {
                                        if (!v[wx, wy, wz])
                                        {
                                            isCube = false;
                                        }
                                    }
                                }
                            }
                            if (isCube)
                            {
                                r++;
                            }
                        } while (isCube && x + r < d);
                        --r; // assume that you have a cube that satisfied r=1, that means it went on to r=2,
                        // but then it failed r=2, so go back to r=1.

                        // Add cube to list with largest value of r used
                        box myCube;
                        myCube.x = x;
                        myCube.y = y;
                        myCube.z = z;
                        myCube.r = r;
                        optimizedBoxes.Add(myCube);

                        // Now remove all the points that exist within our cube.
                        for (int wx = x; wx <= r; wx++)
                        {
                            for (int wy = y; wy <= r; wy++)
                            {
                                for (int wz = z; wz <= r; wz++)
                                {
                                    v[wx, wy, wz] = false;
                                }
                            }
                        }
                    }
                }
            }

            foreach (box b in optimizedBoxes)
            {
                writeCube(fout, b.x, b.y, b.z, b.r);
            }

            fout.Close();
        }

        /* This is my convention for writing cubes:
         * Bottom of cube: 0 1 2 3 (where 0 is the (0,0,0) position)
         * Top of cube:    7 6 5 4 (where 5 is the (1,1,1) position)
         */
        static void writeCube(StreamWriter sw, int x, int y, int z, int r)
        {
            sw.WriteLine("v {0} {1} {2}", x, y, z);
            sw.WriteLine("v {0} {1} {2}", x + r, y, z);
            sw.WriteLine("v {0} {1} {2}", x + r, y + r, z);
            sw.WriteLine("v {0} {1} {2}", x, y + r, z);
            sw.WriteLine("v {0} {1} {2}", x + r, y + r, z + 1);
            sw.WriteLine("v {0} {1} {2}", x + r, y, z + r);
            sw.WriteLine("v {0} {1} {2}", x, y, z + r);
            sw.WriteLine("v {0} {1} {2}", x, y + r, z + r);

            // When building faces remember that faces have direction.
            // Use the right-hand rule to figure out what direction that is.
            sw.WriteLine("f {0}// {1}// {2}// {3}//", totalVerticesWritten + 3, totalVerticesWritten + 2,
                totalVerticesWritten + 1, totalVerticesWritten);
            sw.WriteLine("f {0}// {1}// {2}// {3}//", totalVerticesWritten + 7, totalVerticesWritten + 6,
                totalVerticesWritten + 5, totalVerticesWritten + 4);
            sw.WriteLine("f {0}// {1}// {2}// {3}//", totalVerticesWritten, totalVerticesWritten + 1,
                totalVerticesWritten + 6, totalVerticesWritten + 7);
            sw.WriteLine("f {0}// {1}// {2}// {3}//", totalVerticesWritten + 3, totalVerticesWritten,
                totalVerticesWritten + 7, totalVerticesWritten + 4);
            sw.WriteLine("f {0}// {1}// {2}// {3}//", totalVerticesWritten + 2, totalVerticesWritten + 3,
                totalVerticesWritten + 4, totalVerticesWritten + 5);
            sw.WriteLine("f {0}// {1}// {2}// {3}//", totalVerticesWritten + 1, totalVerticesWritten + 2,
                totalVerticesWritten + 5, totalVerticesWritten + 6);

            totalVerticesWritten += 8;
        }

        static void writeVertex(StreamWriter sw, int x, int y, int z)
        {
            sw.WriteLine("v {0} {1} {2}", x, y, z);
        }
    }
}
